@@ -10,7 +10,7 @@ BEGIN {
 
 BEGIN {
 	$Type::Params::AUTHORITY = 'cpan:TOBYINK';
-	$Type::Params::VERSION   = '1.012005';
+	$Type::Params::VERSION   = '1.016006';
 }
 
 $Type::Params::VERSION =~ tr/_//d;
@@ -31,13 +31,6 @@ our @EXPORT = qw( compile compile_named );
 our @EXPORT_OK =
 	qw( multisig validate validate_named compile_named_oo Invocant wrap_subs wrap_methods ArgsObject );
 	
-sub english_list {
-	require Type::Utils;
-	goto \&Type::Utils::english_list;
-}
-
-my $QUOTE = \&B::perlstring;
-
 {
 	my $Invocant;
 	
@@ -70,7 +63,7 @@ my $QUOTE = \&B::perlstring;
 				},
 				inline_generator => sub {
 					my $param  = shift;
-					my $quoted = $QUOTE->( $param );
+					my $quoted = B::perlstring( $param );
 					sub {
 						my $var = pop;
 						return (
@@ -97,771 +90,41 @@ my $QUOTE = \&B::perlstring;
 	}
 }
 
-sub _mkslurpy {
-	my ( $name, $type, $tc, $i ) = @_;
-	$name = 'local $_' if $name eq '$_';
-	
-	$type eq '@'
-		? sprintf(
-		'%s = [ @_[%d..$#_] ];',
-		$name,
-		$i,
-		)
-		: sprintf(
-		'%s = (@_==%d and ref $_[%d] eq "HASH") ? +{ %%{$_[%d]} } : (($#_-%d)%%2)==0 ? "Error::TypeTiny::WrongNumberOfParameters"->throw(message => sprintf("Odd number of elements in %%s", %s)) : +{ @_[%d..$#_] };',
-		$name,
-		$i + 1,
-		$i,
-		$i,
-		$i,
-		$QUOTE->( "$tc" ),
-		$i,
-		);
-} #/ sub _mkslurpy
-
-sub _mkdefault {
-	my $param_options = shift;
-	my $default;
-	
-	if ( exists $param_options->{default} ) {
-		$default = $param_options->{default};
-		if ( Types::Standard::is_ArrayRef( $default ) and not @$default ) {
-			$default = '[]';
-		}
-		elsif ( Types::Standard::is_HashRef( $default ) and not %$default ) {
-			$default = '{}';
-		}
-		elsif ( Types::Standard::is_Str( $default ) ) {
-			$default = $QUOTE->( $default );
-		}
-		elsif ( Types::Standard::is_Undef( $default ) ) {
-			$default = 'undef';
-		}
-		elsif ( not Types::TypeTiny::is_CodeLike( $default ) ) {
-			Error::TypeTiny::croak(
-				"Default expected to be string, coderef, undef, or reference to an empty hash or array"
-			);
-		}
-	} #/ if ( exists $param_options...)
-	
-	$default;
-} #/ sub _mkdefault
-
-sub _deal_with_head_and_tail {
-	my $options = shift;
-	$options->{arg_fudge_factor}      = 0;
-	$options->{arg_fudge_factor_head} = 0;
-	my @lines;
-	my %env;
-	
-	for my $position ( qw/ head tail / ) {
-		next unless defined $options->{$position};
-		
-		$options->{$position} =
-			[ ( Types::Standard::Any ) x ( 0+ $options->{$position} ) ]
-			if !ref $options->{$position};
-			
-		my $count = @{ $options->{$position} };
-		$options->{arg_fudge_factor}      += $count;
-		$options->{arg_fudge_factor_head} += $count if $position eq 'head';
-		
-		push @lines => (
-			$position eq 'head'
-			? "\@head = splice(\@_, 0, $count);"
-			: "\@tail = splice(\@_, -$count);",
-		);
-		
-		for my $i ( 0 .. $count - 1 ) {
-			my $constraint = $options->{$position}[$i];
-			$constraint = $options->{$position}[$i] = Types::Standard::Any
-				if !ref $constraint && $constraint eq 1;
-			Types::TypeTiny::assert_TypeTiny( $constraint );
-			
-			my $is_optional = 0;
-			$is_optional += grep $_->{uniq} == Types::Standard::Optional->{uniq},
-				$constraint->parents;
-			Error::TypeTiny::croak( "The $position may not contain optional parameters" )
-				if $is_optional;
-				
-			my $varname = sprintf( '$%s[%d]', $position, $i );
-			my $display_var =
-				$position eq 'head'
-				? sprintf( '$_[%d]', $i )
-				: sprintf( '$_[%d]', $i - $count );
-				
-			if ( $constraint->has_coercion and $constraint->coercion->can_be_inlined ) {
-				push @lines, sprintf(
-					'%s = do { %s };',
-					$varname,
-					$constraint->coercion->inline_coercion( $varname ),
-				);
-			}
-			elsif ( $constraint->has_coercion ) {
-				$env{ '@coerce_' . $position }[$i] = $constraint->coercion->compiled_coercion;
-				push @lines, sprintf(
-					'%s = $coerce_%s[%d]->(%s);',
-					$varname,
-					$position,
-					$i,
-					$varname,
-				);
-			} #/ elsif ( $constraint->has_coercion)
-			
-			undef $Type::Tiny::ALL_TYPES{ $constraint->{uniq} };
-			$Type::Tiny::ALL_TYPES{ $constraint->{uniq} } = $constraint;
-			
-			if ( $constraint == Types::Standard::Any ) {
-			
-				# don't really need to check!
-			}
-			elsif ( $constraint->can_be_inlined ) {
-				push @lines, sprintf(
-					'(%s) or Type::Tiny::_failed_check(%d, %s, %s, varname => %s);',
-					$constraint->inline_check( $varname ),
-					$constraint->{uniq},
-					$QUOTE->( $constraint ),
-					$varname,
-					$QUOTE->( $display_var ),
-				);
-			} #/ elsif ( $constraint->can_be_inlined)
-			else {
-				$env{ '@check_' . $position }[$i] = $constraint->compiled_check;
-				push @lines, sprintf(
-					'%s or Type::Tiny::_failed_check(%d, %s, %s, varname => %s);',
-					sprintf( sprintf '$check_%s[%d]->(%s)', $position, $i, $varname ),
-					$constraint->{uniq},
-					$QUOTE->( $constraint ),
-					$varname,
-					$QUOTE->( $display_var ),
-				);
-			} #/ else [ if ( $constraint == Types::Standard::Any)]
-		} #/ for my $i ( 0 .. $count...)
-	} #/ for my $position ( qw/ head tail /)
-	
-	if ( @lines ) {
-		unshift @lines => sprintf(
-			'"Error::TypeTiny::WrongNumberOfParameters"->throw("Not enough parameters to satisfy required head and tail of parameter list") if @_ < %d;',
-			$options->{arg_fudge_factor},
-		);
-		unshift @lines, 'my (@head, @tail);';
-	}
-	
-	\%env, @lines;
-} #/ sub _deal_with_head_and_tail
-
 sub compile {
-	my ( @code, %env );
-	
-	push @code, '#placeholder', '#placeholder';    # @code[0,1]
-	
-	my %options;
-	while ( ref( $_[0] ) eq "HASH" && !$_[0]{slurpy} ) {
-		%options = ( %options, %{ +shift } );
-	}
-	
-	my $arg        = -1;
-	my $saw_slurpy = 0;
-	my $min_args   = 0;
-	my $max_args   = 0;
-	my $saw_opt    = 0;
-	
-	my $return_list = '@_';
-	$code[0] = 'my (%tmp, $tmp);';
-	PARAM: for my $param ( @_ ) {
-		if ( Types::Standard::is_HashRef( $param ) ) {
-			$code[0] = 'my (@R, %tmp, $tmp, $dtmp);';
-			$return_list = '@R';
-			last PARAM;
-		}
-		elsif ( not Types::Standard::is_Bool( $param ) ) {
-			if ( $param->has_coercion ) {
-				$code[0] = 'my (@R, %tmp, $tmp, $dtmp);';
-				$return_list = '@R';
-				last PARAM;
-			}
-		}
-	} #/ PARAM: for my $param ( @_ )
-	
-	my ( $extra_env, @extra_lines ) = _deal_with_head_and_tail( \%options );
-	if ( @extra_lines ) {
-		push @code, @extra_lines;
-		%env         = ( %$extra_env, %env );
-		$return_list = '(@head, @R, @tail)';
-	}
-	
-	my $for =
-		$options{subname}
-		|| [ caller( 1 + ( $options{caller_level} || 0 ) ) ]->[3]
-		|| '__ANON__';
-		
-	my @default_indices;
-	my @default_values;
-	
-	while ( @_ ) {
-		++$arg;
-		my $constraint = shift;
-		my $is_optional;
-		my $really_optional;
-		my $is_slurpy;
-		my $varname;
-		
-		my $param_options = {};
-		$param_options = shift
-			if Types::Standard::is_HashRef( $_[0] ) && !exists $_[0]{slurpy};
-		my $default = _mkdefault( $param_options );
-		
-		if ( $param_options->{optional} or defined $default ) {
-			$is_optional = 1;
-		}
-		
-		if ( Types::Standard::is_Bool( $constraint ) ) {
-			$constraint =
-				$constraint
-				? Types::Standard::Any
-				: Types::Standard::Optional [Types::Standard::Any];
-		}
-		
-		if ( Types::Standard::is_HashRef( $constraint )
-			and exists $constraint->{slurpy} )
-		{
-			$constraint = Types::TypeTiny::to_TypeTiny(
-				$constraint->{slurpy}
-					or Error::TypeTiny::croak( "Slurpy parameter malformed" )
-			);
-			push @code,
-				$constraint->is_a_type_of( Types::Standard::Dict )
-				? _mkslurpy( '$_', '%', $constraint => $arg )
-				: $constraint->is_a_type_of( Types::Standard::Map )
-				? _mkslurpy( '$_', '%', $constraint => $arg )
-				: $constraint->is_a_type_of( Types::Standard::Tuple )
-				? _mkslurpy( '$_', '@', $constraint => $arg )
-				: $constraint->is_a_type_of( Types::Standard::HashRef )
-				? _mkslurpy( '$_', '%', $constraint => $arg )
-				: $constraint->is_a_type_of( Types::Standard::ArrayRef )
-				? _mkslurpy( '$_', '@', $constraint => $arg )
-				: Error::TypeTiny::croak(
-				"Slurpy parameter not of type HashRef or ArrayRef" );
-			$varname = '$_';
-			$is_slurpy++;
-			$saw_slurpy++;
-		} #/ if ( Types::Standard::is_HashRef...)
-		else {
-			Error::TypeTiny::croak( "Parameter following slurpy parameter" ) if $saw_slurpy;
-			
-			$is_optional += grep $_->{uniq} == Types::Standard::Optional->{uniq},
-				$constraint->parents;
-			$really_optional =
-				$is_optional
-				&& $constraint->parent
-				&& $constraint->parent->{uniq} eq Types::Standard::Optional->{uniq}
-				&& $constraint->type_parameter;
-				
-			if ( ref $default ) {
-				$env{'@default'}[$arg] = $default;
-				push @code, sprintf(
-					'$dtmp = ($#_ < %d) ? $default[%d]->() : $_[%d];',
-					$arg,
-					$arg,
-					$arg,
-				);
-				$saw_opt++;
-				$max_args++;
-				$varname = '$dtmp';
-			} #/ if ( ref $default )
-			elsif ( defined $default ) {
-				push @code, sprintf(
-					'$dtmp = ($#_ < %d) ? %s : $_[%d];',
-					$arg,
-					$default,
-					$arg,
-				);
-				$saw_opt++;
-				$max_args++;
-				$varname = '$dtmp';
-			} #/ elsif ( defined $default )
-			elsif ( $is_optional ) {
-				push @code, sprintf(
-					'return %s if $#_ < %d;',
-					$return_list,
-					$arg,
-				);
-				$saw_opt++;
-				$max_args++;
-				$varname = sprintf '$_[%d]', $arg;
-			} #/ elsif ( $is_optional )
-			else {
-				Error::TypeTiny::croak( "Non-Optional parameter following Optional parameter" )
-					if $saw_opt;
-				$min_args++;
-				$max_args++;
-				$varname = sprintf '$_[%d]', $arg;
-			}
-		} #/ else [ if ( Types::Standard::is_HashRef...)]
-		
-		if ( $constraint->has_coercion and $constraint->coercion->can_be_inlined ) {
-			push @code, sprintf(
-				'$tmp%s = %s;',
-				( $is_optional ? '{x}' : '' ),
-				$constraint->coercion->inline_coercion( $varname )
-			);
-			$varname = '$tmp' . ( $is_optional ? '{x}' : '' );
-		}
-		elsif ( $constraint->has_coercion ) {
-			$env{'@coerce'}[$arg] = $constraint->coercion->compiled_coercion;
-			push @code, sprintf(
-				'$tmp%s = $coerce[%d]->(%s);',
-				( $is_optional ? '{x}' : '' ),
-				$arg,
-				$varname,
-			);
-			$varname = '$tmp' . ( $is_optional ? '{x}' : '' );
-		} #/ elsif ( $constraint->has_coercion)
-		
-		# unweaken the constraint in the cache
-		undef $Type::Tiny::ALL_TYPES{ $constraint->{uniq} };
-		$Type::Tiny::ALL_TYPES{ $constraint->{uniq} } = $constraint;
-		
-		if ( $constraint->can_be_inlined ) {
-			push @code, sprintf(
-				'(%s) or Type::Tiny::_failed_check(%d, %s, %s, varname => %s);',
-				$really_optional
-				? $constraint->type_parameter->inline_check( $varname )
-				: $constraint->inline_check( $varname ),
-				$constraint->{uniq},
-				$QUOTE->( $constraint ),
-				$varname,
-				$is_slurpy
-				? 'q{$SLURPY}'
-				: sprintf( 'q{$_[%d]}', $arg + $options{arg_fudge_factor_head} ),
-			);
-		} #/ if ( $constraint->can_be_inlined)
-		else {
-			$env{'@check'}[$arg] =
-				$really_optional
-				? $constraint->type_parameter->compiled_check
-				: $constraint->compiled_check;
-			push @code, sprintf(
-				'%s or Type::Tiny::_failed_check(%d, %s, %s, varname => %s);',
-				sprintf( sprintf '$check[%d]->(%s)', $arg, $varname ),
-				$constraint->{uniq},
-				$QUOTE->( $constraint ),
-				$varname,
-				$is_slurpy
-				? 'q{$SLURPY}'
-				: sprintf( 'q{$_[%d]}', $arg + $options{arg_fudge_factor_head} ),
-			);
-		} #/ else [ if ( $constraint->can_be_inlined)]
-		
-		unless ( $return_list eq '@_' ) {
-			push @code, sprintf 'push @R, %s;', $varname;
-		}
-	} #/ while ( @_ )
-	
-	my $thang = 'scalar(@_)';
-	
-	if ( $min_args == $max_args and not $saw_slurpy ) {
-		$code[1] = sprintf(
-			'"Error::TypeTiny::WrongNumberOfParameters"->throw(got => %s, minimum => %d, maximum => %d) if @_ != %d;',
-			$thang,
-			$min_args + $options{arg_fudge_factor},
-			$max_args + $options{arg_fudge_factor},
-			$min_args + $options{arg_fudge_factor},
-		);
-	}
-	elsif ( $min_args < $max_args and not $saw_slurpy ) {
-		$code[1] = sprintf(
-			'"Error::TypeTiny::WrongNumberOfParameters"->throw(got => %s, minimum => %d, maximum => %d) if @_ < %d || @_ > %d;',
-			$thang,
-			$min_args + $options{arg_fudge_factor},
-			$max_args + $options{arg_fudge_factor},
-			$min_args + $options{arg_fudge_factor},
-			$max_args + $options{arg_fudge_factor},
-		);
-	} #/ elsif ( $min_args < $max_args...)
-	elsif ( $min_args and $saw_slurpy ) {
-		$code[1] = sprintf(
-			'"Error::TypeTiny::WrongNumberOfParameters"->throw(got => %s, minimum => %d) if @_ < %d;',
-			$thang,
-			$min_args + $options{arg_fudge_factor},
-			$min_args + $options{arg_fudge_factor},
-		);
-	}
-	
-	push @code, $return_list;
-	
-	my $source = "sub { no warnings; " . join( "\n", @code ) . " };";
-	
-	return $source if $options{want_source};
-	
-	my $closure = eval_closure(
-		source      => $source,
-		description => $options{description}
-			|| sprintf( "parameter validation for '%s'", $for ),
-		environment => \%env,
-	);
-	
-	return {
-		min_args => $options{arg_fudge_factor} + ( $min_args || 0 ),
-		max_args => $saw_slurpy
-		? undef
-		: $options{arg_fudge_factor} + ( $max_args || 0 ),
-		closure     => $closure,
-		source      => $source,
-		environment => \%env,
-	} if $options{want_details};
-	
-	return $closure;
-} #/ sub compile
+	require Type::Params::Signature;
+
+	my $sig = 'Type::Params::Signature'->new_from_compile( pos => @_ );
+	my $for = [ caller( 1 + ( $sig->{caller_level} || 0 ) ) ]->[3] || '__ANON__::__ANON__';
+	my ( $pkg, $sub ) = ( $for =~ /^(.+)::(\w+)$/ );
+	$sig->{package} ||= $pkg;
+	$sig->{subname} ||= $sub;
+
+	$sig->return_wanted;
+}
 
 sub compile_named {
-	my ( @code, %env );
-	
-	push @code, 'my (%R, %tmp, $tmp);';
-	push @code, '#placeholder';           # $code[1]
-	
-	my %options;
-	while ( ref( $_[0] ) eq "HASH" && !$_[0]{slurpy} ) {
-		%options = ( %options, %{ +shift } );
-	}
-	my $arg = -1;
-	my $had_slurpy;
-	
-	my ( $extra_env, @extra_lines ) = _deal_with_head_and_tail( \%options );
-	if ( @extra_lines ) {
-		push @code, @extra_lines;
-		%env = ( %$extra_env, %env );
-	}
-	
-	my $for =
-		$options{subname}
-		|| [ caller( 1 + ( $options{caller_level} || 0 ) ) ]->[3]
-		|| '__ANON__';
-		
-	push @code,
-		'my %in = ((@_==1) && ref($_[0]) eq "HASH") ? %{$_[0]} : (@_ % 2) ? "Error::TypeTiny::WrongNumberOfParameters"->throw(message => "Odd number of elements in hash") : @_;';
-	my @names;
-	
-	while ( @_ ) {
-		++$arg;
-		my ( $name, $constraint ) = splice( @_, 0, 2 );
-		push @names, $name;
-		
-		my $is_optional;
-		my $really_optional;
-		my $is_slurpy;
-		my $varname;
-		my $default;
-		
-		Types::Standard::is_Str( $name )
-			or Error::TypeTiny::croak( "Expected parameter name as string, got $name" );
-			
-		my $param_options = {};
-		$param_options = shift @_
-			if Types::Standard::is_HashRef( $_[0] ) && !exists $_[0]{slurpy};
-		$default = _mkdefault( $param_options );
-		
-		if ( $param_options->{optional} or defined $default ) {
-			$is_optional = 1;
-		}
-		
-		if ( Types::Standard::is_Bool( $constraint ) ) {
-			$constraint =
-				$constraint
-				? Types::Standard::Any
-				: Types::Standard::Optional [Types::Standard::Any];
-		}
-		
-		if ( Types::Standard::is_HashRef( $constraint )
-			and exists $constraint->{slurpy} )
-		{
-			$constraint = Types::TypeTiny::to_TypeTiny( $constraint->{slurpy} );
-			++$is_slurpy;
-			++$had_slurpy;
-		}
-		else {
-			$is_optional += grep $_->{uniq} == Types::Standard::Optional->{uniq},
-				$constraint->parents;
-			$really_optional =
-				$is_optional
-				&& $constraint->parent
-				&& $constraint->parent->{uniq} eq Types::Standard::Optional->{uniq}
-				&& $constraint->type_parameter;
-				
-			$constraint = $constraint->type_parameter if $really_optional;
-		} #/ else [ if ( Types::Standard::is_HashRef...)]
-		
-		if ( ref $default ) {
-			$env{'@default'}[$arg] = $default;
-			push @code, sprintf(
-				'exists($in{%s}) or $in{%s} = $default[%d]->();',
-				$QUOTE->( $name ),
-				$QUOTE->( $name ),
-				$arg,
-			);
-		}
-		elsif ( defined $default ) {
-			push @code, sprintf(
-				'exists($in{%s}) or $in{%s} = %s;',
-				$QUOTE->( $name ),
-				$QUOTE->( $name ),
-				$default,
-			);
-		}
-		elsif ( not $is_optional || $is_slurpy ) {
-			push @code, sprintf(
-				'exists($in{%s}) or "Error::TypeTiny::WrongNumberOfParameters"->throw(message => sprintf "Missing required parameter: %%s", %s);',
-				$QUOTE->( $name ),
-				$QUOTE->( $name ),
-			);
-		}
-		
-		my $need_to_close_if = 0;
-		
-		if ( $is_slurpy ) {
-			$varname = '\\%in';
-		}
-		elsif ( $is_optional ) {
-			push @code, sprintf( 'if (exists($in{%s})) {',  $QUOTE->( $name ) );
-			push @code, sprintf( '$tmp = delete($in{%s});', $QUOTE->( $name ) );
-			$varname = '$tmp';
-			++$need_to_close_if;
-		}
-		else {
-			push @code, sprintf( '$tmp = delete($in{%s});', $QUOTE->( $name ) );
-			$varname = '$tmp';
-		}
-		
-		if ( $constraint->has_coercion ) {
-			if ( $constraint->coercion->can_be_inlined ) {
-				push @code, sprintf(
-					'$tmp = %s;',
-					$constraint->coercion->inline_coercion( $varname )
-				);
-			}
-			else {
-				$env{'@coerce'}[$arg] = $constraint->coercion->compiled_coercion;
-				push @code, sprintf(
-					'$tmp = $coerce[%d]->(%s);',
-					$arg,
-					$varname,
-				);
-			}
-			$varname = '$tmp';
-		} #/ if ( $constraint->has_coercion)
-		
-		if ( $constraint->can_be_inlined ) {
-			push @code, sprintf(
-				'(%s) or Type::Tiny::_failed_check(%d, %s, %s, varname => %s);',
-				$constraint->inline_check( $varname ),
-				$constraint->{uniq},
-				$QUOTE->( $constraint ),
-				$varname,
-				$is_slurpy ? 'q{$SLURPY}' : sprintf( 'q{$_{%s}}', $QUOTE->( $name ) ),
-			);
-		} #/ if ( $constraint->can_be_inlined)
-		else {
-			$env{'@check'}[$arg] = $constraint->compiled_check;
-			push @code, sprintf(
-				'%s or Type::Tiny::_failed_check(%d, %s, %s, varname => %s);',
-				sprintf( sprintf '$check[%d]->(%s)', $arg, $varname ),
-				$constraint->{uniq},
-				$QUOTE->( $constraint ),
-				$varname,
-				$is_slurpy ? 'q{$SLURPY}' : sprintf( 'q{$_{%s}}', $QUOTE->( $name ) ),
-			);
-		} #/ else [ if ( $constraint->can_be_inlined)]
-		
-		push @code, sprintf( '$R{%s} = %s;', $QUOTE->( $name ), $varname );
-		
-		push @code, '}' if $need_to_close_if;
-	} #/ while ( @_ )
-	
-	if ( !$had_slurpy ) {
-		push @code,
-			'keys(%in) and "Error::TypeTiny"->throw(message => sprintf "Unrecognized parameter%s: %s", keys(%in)>1?"s":"", Type::Params::english_list(sort keys %in));';
-	}
-	
-	if ( $options{named_to_list} ) {
-		Error::TypeTiny::croak( "named_to_list option cannot be used with slurpy" )
-			if $had_slurpy;
-		my @order = ref $options{named_to_list} ? @{ $options{named_to_list} } : @names;
-		push @code, sprintf( '@R{%s}', join ",", map $QUOTE->( $_ ), @order );
-	}
-	elsif ( $options{bless} ) {
-		if ( $options{oo_trace} ) {
-			push @code, sprintf( '$R{%s} = %s;', $QUOTE->( '~~caller' ), $QUOTE->( $for ) );
-		}
-		push @code, sprintf( 'bless \\%%R, %s;', $QUOTE->( $options{bless} ) );
-	}
-	elsif ( Types::Standard::is_ArrayRef( $options{class} ) ) {
-		push @code,
-			sprintf(
-			'(%s)->%s(\\%%R);', $QUOTE->( $options{class}[0] ),
-			$options{class}[1] || 'new'
-			);
-	}
-	elsif ( $options{class} ) {
-		push @code,
-			sprintf(
-			'(%s)->%s(\\%%R);', $QUOTE->( $options{class} ),
-			$options{constructor} || 'new'
-			);
-	}
-	else {
-		push @code, '\\%R;';
-	}
-	
-	if ( $options{head} || $options{tail} ) {
-		$code[-1] = 'my @R = ' . $code[-1] . ';';
-		push @code, 'unshift @R, @head;' if $options{head};
-		push @code, 'push @R, @tail;'    if $options{tail};
-		push @code, '@R;';
-	}
-	
-	my $source = "sub { no warnings; " . join( "\n", @code ) . " };";
-	return $source if $options{want_source};
-	
-	my $closure = eval_closure(
-		source      => $source,
-		description => $options{description}
-			|| sprintf( "parameter validation for '%s'", $for ),
-		environment => \%env,
-	);
-	
-	my $max_args = undef;
-	if ( !$had_slurpy ) {
-		$max_args = 2 * ( $arg + 1 );
-		$max_args += $options{arg_fudge_factor};
-	}
-	
-	return {
-		min_args    => $options{arg_fudge_factor},
-		max_args    => $max_args,
-		closure     => $closure,
-		source      => $source,
-		environment => \%env,
-	} if $options{want_details};
-	
-	return $closure;
-} #/ sub compile_named
+	require Type::Params::Signature;
 
-my %klasses;
-my $kls_id = 0;
-my $has_cxsa;
-my $want_cxsa;
+	my $sig = 'Type::Params::Signature'->new_from_compile( named => @_ );
+	my $for = [ caller( 1 + ( $sig->{caller_level} || 0 ) ) ]->[3] || '__ANON__::__ANON__';
+	my ( $pkg, $sub ) = ( $for =~ /^(.+)::(\w+)$/ );
+	$sig->{package} ||= $pkg;
+	$sig->{subname} ||= $sub;
 
-sub _mkklass {
-	my $klass = sprintf( '%s::OO::Klass%d', __PACKAGE__, ++$kls_id );
-	
-	if ( !defined $has_cxsa or !defined $want_cxsa ) {
-		$has_cxsa = !!eval {
-			require Class::XSAccessor;
-			'Class::XSAccessor'->VERSION( '1.17' );    # exists_predicates, June 2013
-			1;
-		};
-		
-		$want_cxsa =
-			$ENV{PERL_TYPE_PARAMS_XS}             ? 'XS'
-			: exists( $ENV{PERL_TYPE_PARAMS_XS} ) ? 'PP'
-			: $has_cxsa                           ? 'XS'
-			:                                       'PP';
-			
-		if ( $want_cxsa eq 'XS' and not $has_cxsa ) {
-			Error::TypeTiny::croak( "Cannot load Class::XSAccessor" );    # uncoverable statement
-		}
-	} #/ if ( !defined $has_cxsa...)
-	
-	if ( $want_cxsa eq 'XS' ) {
-		eval {
-			'Class::XSAccessor'->import(
-				redefine => 1,
-				class    => $klass,
-				getters  => {
-					map { defined( $_->{getter} ) ? ( $_->{getter} => $_->{slot} ) : () }
-						values %{ $_[0] }
-				},
-				exists_predicates => {
-					map { defined( $_->{predicate} ) ? ( $_->{predicate} => $_->{slot} ) : () }
-						values %{ $_[0] }
-				},
-			);
-			1;
-		} ? return ( $klass ) : die( $@ );
-	} #/ if ( $want_cxsa eq 'XS')
-	
-	for my $attr ( values %{ $_[0] } ) {
-		defined( $attr->{getter} ) and eval sprintf(
-			'package %s; sub %s { $_[0]{%s} }; 1',
-			$klass,
-			$attr->{getter},
-			$attr->{slot},
-		) || die( $@ );
-		defined( $attr->{predicate} ) and eval sprintf(
-			'package %s; sub %s { exists $_[0]{%s} }; 1',
-			$klass,
-			$attr->{predicate},
-			$attr->{slot},
-		) || die( $@ );
-	} #/ for my $attr ( values %...)
-	
-	$klass;
-} #/ sub _mkklass
+	$sig->return_wanted;
+}
 
 sub compile_named_oo {
-	my %options;
-	while ( ref( $_[0] ) eq "HASH" && !$_[0]{slurpy} ) {
-		%options = ( %options, %{ +shift } );
-	}
-	my @rest = @_;
-	
-	my %attribs;
-	while ( @_ ) {
-		my ( $name, $type ) = splice( @_, 0, 2 );
-		my $opts =
-			( Types::Standard::is_HashRef( $_[0] ) && !exists $_[0]{slurpy} )
-			? shift( @_ )
-			: {};
-			
-		my $is_optional = 0+ !!$opts->{optional};
-		$is_optional += grep $_->{uniq} == Types::Standard::Optional->{uniq},
-			$type->parents;
-			
-		my $getter =
-			exists( $opts->{getter} )
-			? $opts->{getter}
-			: $name;
-			
-		Error::TypeTiny::croak( "Bad accessor name: $getter" )
-			unless $getter =~ /\A[A-Za-z][A-Za-z0-9_]*\z/;
-			
-		my $predicate =
-			exists( $opts->{predicate} )
-			? (
-			$opts->{predicate} eq '1'   ? "has_$getter"
-			: $opts->{predicate} eq '0' ? undef
-			:                             $opts->{predicate}
-			)
-			: ( $is_optional ? "has_$getter" : undef );
-			
-		$attribs{$name} = {
-			slot      => $name,
-			getter    => $getter,
-			predicate => $predicate,
-		};
-	} #/ while ( @_ )
-	
-	my $kls = join '//',
-		map sprintf(
-		'%s*%s*%s', $attribs{$_}{slot}, $attribs{$_}{getter},
-		$attribs{$_}{predicate} || '0'
-		),
-		sort keys %attribs;
-		
-	$klasses{$kls} ||= _mkklass( \%attribs );
-	
-	@_ = ( { oo_trace => 1, %options, bless => $klasses{$kls} }, @rest );
-	goto \&compile_named;
-} #/ sub compile_named_oo
+	require Type::Params::Signature;
+
+	my $sig = 'Type::Params::Signature'->new_from_compile( named => { bless => 1 }, @_ );
+	my $for = [ caller( 1 + ( $sig->{caller_level} || 0 ) ) ]->[3] || '__ANON__::__ANON__';
+	my ( $pkg, $sub ) = ( $for =~ /^(.+)::(\w+)$/ );
+	$sig->{package} ||= $pkg;
+	$sig->{subname} ||= $sub;
+
+	$sig->return_wanted;
+}
 
 # Would be faster to inline this into validate and validate_named, but
 # that would complicate them. :/
@@ -876,7 +139,7 @@ sub _mk_key {
 			: Types::TypeTiny::is_TypeTiny( $_ ) ? sprintf( 'TYPE=%s', $_->{uniq} )
 			: Types::Standard::is_Ref( $_ )      ? sprintf( 'REF=%s', refaddr( $_ ) )
 			: Types::Standard::is_Undef( $_ )    ? sprintf( 'UNDEF' )
-			: $QUOTE->( $_ )
+			: B::perlstring( $_ )
 	} @_;
 } #/ sub _mk_key
 
@@ -908,8 +171,16 @@ sub validate_named {
 	goto $sub;
 } #/ sub validate_named
 
+sub _deal_with_on_die {
+	my $options = shift;
+	if ( $options->{'on_die'} ) {
+		return ( { '$on_die' => \$options->{'on_die'} }, '1;' );
+	}
+	return ( {}, 'my $on_die = undef;' );
+}
+
 sub multisig {
-	my %options = ( ref( $_[0] ) eq "HASH" && !$_[0]{slurpy} ) ? %{ +shift } : ();
+	my %options = ( ref( $_[0] ) eq "HASH" ) ? %{ +shift } : ();
 	$options{message}     ||= "Parameter validation failed";
 	$options{description} ||= sprintf(
 		"parameter validation for '%s'",
@@ -926,9 +197,18 @@ sub multisig {
 			: Types::TypeTiny::is_ArrayLike( $_ ) ? compile( { want_details => 1 }, @$_ )
 			:                                       $_;
 	} @_;
-	
+
+	my %env;
 	my @code = 'sub { my $r; ';
 	
+	{
+		my ( $extra_env, @extra_lines ) = _deal_with_on_die( \%options );
+		if ( @extra_lines ) {
+			$code[0] .= join '', @extra_lines;
+			%env = ( %$extra_env, %env );
+		}
+	}
+	
 	for my $i ( 0 .. $#multi ) {
 		my $flag = sprintf( '${^TYPE_PARAMS_MULTISIG} = %d', $i );
 		my $sig  = $multi[$i];
@@ -951,7 +231,7 @@ sub multisig {
 	
 	push @code,
 		sprintf(
-		'"Error::TypeTiny"->throw(message => "%s");',
+		'return "Error::TypeTiny"->throw_cb($on_die, message => "%s");',
 		quotemeta( "$options{message}" )
 		);
 	push @code, '}';
@@ -959,7 +239,7 @@ sub multisig {
 	eval_closure(
 		source      => \@code,
 		description => $options{description},
-		environment => { '@multi' => \@multi },
+		environment => { '@multi' => \@multi, %env },
 	);
 } #/ sub multisig
 
@@ -1143,7 +423,7 @@ The generalized form of specifications for positional parameters is:
    $type_for_arg_2, \%opts_for_arg_2,
    $type_for_arg_3, \%opts_for_arg_3,
    ...,
-   slurpy($slurpy_type),
+   Slurpy[...],
  );
 
 If a hashref of options is empty, it can simply be omitted. Much of the
@@ -1221,7 +501,7 @@ some required parameters I<after> a slurpy or optional parameter.
 
   my $check = compile(
     { tail => [ CodeRef ] },
-    slurpy ArrayRef[Str],
+    Slurpy[ ArrayRef[Str] ],
   );
   
   my ($strings, $coderef) = $check->("foo", "bar", sub { ... });
@@ -1252,6 +532,22 @@ use this.
 If you wish to use the default description, but need to change the caller
 level for detecting the sub name, use this.
 
+=item C<< on_die >> B<< Maybe[CodeRef] >>
+
+  my $check = compile(
+    { on_die => sub { ... } },
+    ...,
+  );
+  
+  my @args = $check->( @_ );
+
+Normally, at the first invalid argument the C<< $check >> coderef encounters,
+it will throw an exception.
+
+If an C<on_die> coderef is provided, then it is called instead, and the
+exception is passed to it as an object. The C<< $check >> coderef will still
+immediately return though.
+
 =back
 
 The types for each parameter may be any L<Type::Tiny> type constraint, or
@@ -1308,7 +604,7 @@ This is an alternative way of indicating that a parameter is optional.
 The two are not I<exactly> equivalent. The exceptions thrown will
 differ in the type name they mention. (B<Int> versus B<< Optional[Int] >>.)
 
-=item C<< default >> B<< CodeRef|Ref|Str|Undef >>
+=item C<< default >> B<< CodeRef|ScalarRef|Ref|Str|Undef >>
 
 A default may be provided for a parameter.
 
@@ -1335,6 +631,15 @@ non-empty hashrefs. For undef, simple strings, numbers, and empty
 structures, avoiding using a coderef will make your parameter processing
 faster.
 
+Instead of a coderef, you can use a reference to a string of Perl source
+code:
+
+ state $check = compile(
+   Int,
+   Int, { default => \ '6 * 111' },
+   Int, { default => \ '9 * 111' },
+ );
+
 The default I<will> be validated against the type constraint, and
 potentially coerced.
 
@@ -1343,6 +648,34 @@ used in a particular check, will slightly slow down C<< $check >>
 because it means that C<< $check >> can't just check C<< @_ >> and return
 it unaltered if it's valid — it needs to build a new array to return.
 
+=item C<< clone >> B<Bool>
+
+If this is set to true, it will deep clone incoming values via C<dclone>
+from L<Storable> (a core module since Perl 5.7.3).
+
+In the below example, C<< $arr >> is a reference to a I<clone of>
+C<< @numbers >>, so pushing additional numbers to it leaves C<< @numbers >>
+unaffected.
+
+ sub foo {
+   state $check = compile( ArrayRef, { clone => 1 } );
+   my ( $arr ) = &$check;
+   
+   push @$arr, 4, 5, 6;
+ }
+ 
+ my @numbers = ( 1, 2, 3 );
+ foo( \@numbers );
+ 
+ print "@numbers\n";  ## 1 2 3
+
+=item C<< slurpy >> B<Bool>
+
+The following two should be equivalent:
+
+ my $check = compile( Int, Slurpy[ArrayRef] );
+ my $check = compile( Int, ArrayRef, { slurpy => 1 } );
+
 =back
 
 As a special case, the numbers 0 and 1 may be used as shortcuts for
@@ -1360,7 +693,7 @@ Defaults are not supported for slurpy parameters.
 Example with a slurpy ArrayRef:
 
  sub xyz {
-   state $check = compile(Int, Int, slurpy ArrayRef[Int]);
+   state $check = compile( Int, Int, Slurpy[ ArrayRef[Int] ] );
    my ($foo, $bar, $baz) = $check->(@_);
  }
  
@@ -1373,7 +706,7 @@ Example with a slurpy HashRef:
  my $check = compile(
    Int,
    Optional[Str],
-   slurpy HashRef[Int],
+   Slurpy[ HashRef[Int] ],
  );
  
  my ($x, $y, $z) = $check->(1, "y", foo => 666, bar => 999);
@@ -1386,14 +719,14 @@ but a type does need to inherit from one of those because otherwise
 Type::Params cannot know what kind of structure to slurp the remaining
 arguments into.
 
-B<< slurpy Any >> is also allowed as a special case, and is treated as
-B<< slurpy ArrayRef >>.
+B<< Slurpy[Any] >> is also allowed as a special case, and is treated as
+B<< Slurpy[ArrayRef] >>.
 
 From Type::Params 1.005000 onwards, slurpy hashrefs can be passed in as a
 true hashref (which will be shallow cloned) rather than key-value pairs.
 
  sub xyz {
-   state $check = compile(Int, slurpy HashRef);
+   state $check = compile(Int, Slurpy[HashRef]);
    my ($num, $hr) = $check->(@_);
    ...
  }
@@ -1440,7 +773,7 @@ parameter:
    bar   => $type_for_bar, \%opts_for_bar,
    baz   => $type_for_baz, \%opts_for_baz,
    ...,
-   extra => slurpy($slurpy_type),
+   extra => Slurpy[...],
  );
 
 The C<< $check >> coderef will return a hashref.
@@ -1517,14 +850,14 @@ Slurpy parameters are slurped into a nested hashref.
   my $check = compile(
     foo    => Str,
     bar    => Optional[Str],
-    extra  => slurpy HashRef[Str],
+    extra  => Slurpy[ HashRef[Str] ],
   );
   my $args = $check->(foo => "aaa", quux => "bbb");
   
   print $args->{foo}, "\n";             # aaa
   print $args->{extra}{quux}, "\n";     # bbb
 
-B<< slurpy Any >> is treated as B<< slurpy HashRef >>.
+B<< slurpy[Any] >> is treated as B<< slurpy[HashRef] >>.
 
 The C<head> and C<tail> options are supported. This allows for a
 mixture of positional and named arguments, as long as the positional
@@ -1556,6 +889,36 @@ This can be combined with C<named_to_list>:
   
   say $bar;  # 'y'
 
+There is one additional parameter option supported, in addition to
+the C<optional>, C<default>, C<clone>, and C<slurpy> options already
+supported by positional parameters.
+
+=over
+
+=item C<alias> B<< Str|ArrayRef[Str] >>
+
+A list of alternative names for the parameter, or a single alternative
+name.
+
+  {
+    my $check;
+    sub adder {
+      $check ||= compile_named(
+        first_number   => Int, { alias => [ 'x' ] },
+        second_number  => Int, { alias =>   'y'   },
+      );
+      my ( $arg ) = &$check;
+      return $arg->{first_number} + $arg->{second_number};
+    }
+  }
+  
+  say adder( first_number => 40, second_number => 2 );    # 42
+  say adder( x            => 40, y             => 2 );    # 42
+  say adder( first_number => 40, y             => 2 );    # 42
+  say adder( first_number => 40, x => 1, y => 2 );        # dies!
+
+=back
+
 =head3 C<< validate_named(\@_, @spec) >>
 
 Like C<compile> has C<validate>, C<compile_named> has C<validate_named>.
@@ -1612,7 +975,9 @@ fast).
 An optional parameter C<foo> will also get a nifty C<< $arg->has_foo >>
 predicate method. Yay!
 
-C<compile_named_oo> gives you some extra options for parameters.
+C<compile_named_oo> gives you some extra options for parameters, in
+addition to the C<optional>, C<default>, C<clone>, C<slurpy>, and
+C<alias> options already supported by C<compile_named>.
 
    sub add_contact_to_database {
       state $check = compile_named_oo(
@@ -1633,6 +998,9 @@ C<compile_named_oo> gives you some extra options for parameters.
 The C<getter> option lets you choose the method name for getting the
 argument value.
 
+If the parameter has an alias, this currently I<does not> result in
+additional getters being defined.
+
 =item C<< predicate >> B<< Str >>
 
 The C<predicate> option lets you choose the method name for checking
@@ -1640,6 +1008,9 @@ the existence of an argument. By setting an explicit predicate method
 name, you can force a predicate method to be generated for non-optional
 arguments.
 
+If the parameter has an alias, this currently I<does not> result in
+additional predicate methods being defined.
+
 =back
 
 The objects returned by C<compile_named_oo> are blessed into lightweight
@@ -1702,7 +1073,7 @@ a full example:
          [ Int, ArrayRef ],
          [ Str, HashRef ],
          sub {
-            my ($meth, $obj);
+            my ($meth, $obj) = @_;
             die unless is_Object($obj);
             die unless $obj->can($meth);
             return ($meth, $obj);
@@ -1868,7 +1239,7 @@ Toby Inkster E<lt>tobyink@cpan.orgE<gt>.
 
 =head1 COPYRIGHT AND LICENCE
 
-This software is copyright (c) 2013-2014, 2017-2021 by Toby Inkster.
+This software is copyright (c) 2013-2014, 2017-2022 by Toby Inkster.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
